﻿Module Main

    Sub Main(ByVal Args As String())

        ' Files name checking
        If Args.Length <> 1 Then
            Console.WriteLine("ERROR: One input TS file must be specified as parameter!")
            Return
        End If
        Dim FileName As String = Args(0)
        If IO.File.Exists(FileName) = False Then
            Console.WriteLine("ERROR: Input file [" + FileName + "] doesn't exist!")
            Return
        End If
        If FileName.Remove(0, FileName.Length - 3).ToLower <> ".ts" Then
            Console.WriteLine("ERROR: Input file [" + FileName + "] isn't a .ts file!")
            Return
        End If
        Dim IndexName As String = FileName + "x"

        ' Looop over TS packets
        Console.WriteLine("Starting indexing of [" + FileName + "]")
        Console.WriteLine("Index file created will be [" + IndexName + "]")
        Console.WriteLine("Searching the first PSI-PMT packet...")
        Dim sr As New IO.StreamReader(FileName)
        Dim sw As New IO.StreamWriter(IndexName, False)
        Dim NbRead As Integer = 0
        Dim Packets As Long = 0
        Dim State As Integer = 0 ' 0=Search PMT 1=Search PCR 2=Search PES
        Dim PCR_PID As Integer = 0
        Dim First_PCR As Double = 0
        Dim Last_PCR As Double = 0
        Dim Index_PCR As Double = 0
        Dim PES_Found As Long = 0
        Dim PES_filtered As Long = 0
        Dim StartTime As DateTime = DateTime.Now
        Dim Buffer() As Byte = CType(Array.CreateInstance(GetType(Byte), 188), Byte())
        Do
            NbRead = sr.BaseStream.Read(Buffer, 0, 188)
            If NbRead = 188 Then

                ' Starting TS packet analysis
                Packets += 1

                ' General TS values
                Dim sync_byte As Byte = Buffer(0)
                Dim transport_error_indicator As Byte = (Buffer(1) And CByte(&H80)) >> 7
                Dim payload_unit_start_indicator As Byte = (Buffer(1) And CByte(&H40)) >> 6
                Dim PID As Integer = ((CInt(Buffer(1)) And CInt(&H1F)) << 8) + CInt(Buffer(2))
                Dim transport_scrambling_control As Byte = (Buffer(3) And CByte(&HC0)) >> 6
                Dim adaptation_field_control As Byte = (Buffer(3) And CByte(&H30)) >> 4

                ' Check if the sync byte is present
                If sync_byte <> &H47 Then
                    Console.WriteLine("ERROR: Missing the TS sync byte!")
                    NbRead = 0
                End If

                ' Searching PMT
                If State = 0 Then
                    ' TS values checking
                    If transport_error_indicator = 0 And payload_unit_start_indicator = 1 And PID <> &H1FFF And transport_scrambling_control = 0 And adaptation_field_control = 1 Then
                        ' Checking if it's not a PES - No packet_start_code_prefix
                        If Not (Buffer(4) = 0 And Buffer(5) = 0 And Buffer(6) = 1) Then
                            ' Checking if there is PSI table signature with table_id=2 for PMT and we assume the pointer_field=0 (maybe not always true...)
                            If Buffer(4) = 0 And Buffer(5) = 2 And (Buffer(6) And CByte(&HC0)) = CByte(&H80) Then
                                ' Ok, we have the PMT, we need will get the PCR PID and assume that it is the Video PID
                                Console.WriteLine(" * PMT PID is " + PID.ToString + " (0x" + PID.ToString("X") + ") - " + (Packets - 1).ToString + " Packets skipped since beginning")
                                PCR_PID = ((CInt(Buffer(13)) And CInt(&H1F)) << 8) + CInt(Buffer(14))
                                Console.WriteLine(" * PCR PID is " + PCR_PID.ToString + " (0x" + PCR_PID.ToString("X") + ") - We assume it is also the video PID")
                                ' Next, we search the first PCR
                                State = 1
                                Console.WriteLine("Searching the first PCR...")
                            End If
                        End If
                    End If
                End If

                ' Searching the PCR
                If State = 1 Or State = 2 Then
                    ' Checking some TS values - Looking for PCR PID with adpation field
                    If PID = PCR_PID And transport_error_indicator = 0 And (adaptation_field_control = 2 Or adaptation_field_control = 3) Then
                        ' Checking if adpation field length is not nul and if PCR flag is set
                        If Buffer(4) > 0 And (Buffer(5) And CByte(&H10)) = CByte(&H10) Then
                            ' Getting the raw PCR values
                            Dim PCR_Base As Long = (CLng(Buffer(6)) << CLng(25)) + (CLng(Buffer(7)) << CLng(17)) + (CLng(Buffer(8)) << CLng(9)) + (CLng(Buffer(9)) << CLng(1)) + ((CLng(Buffer(10)) And CLng(&H80)) >> CLng(7))
                            Dim PCR_Ext As Integer = ((CInt(Buffer(10)) And CInt(&H1)) << CInt(8)) + CInt(Buffer(11))
                            ' Transforming to float
                            Last_PCR = (PCR_Base / 90000) + (PCR_Ext / 27000000)
                            ' Check if we were searching the first PCR
                            If State = 1 Then
                                ' We start indexing
                                State = 2
                                First_PCR = Last_PCR
                                Console.WriteLine(" * First PCR value is " + First_PCR.ToString("0.000") + "s - " + (Packets - 1).ToString + " Packets skipped since beginning")
                                Console.WriteLine("Searching PES sections start for indexing...")
                            End If
                        End If
                    End If
                End If

                ' Indexing 
                If State = 2 Then
                    ' Checking some TS values - Looking for PCR PID with start section
                    If PID = PCR_PID And transport_error_indicator = 0 And payload_unit_start_indicator = 1 Then
                        ' We certainly have a PES section start here
                        PES_Found += 1
                        ' Check if we have a new PCR
                        If Index_PCR <> Last_PCR Then
                            Index_PCR = Last_PCR
                            ' We create an index
                            Dim Index() As Byte = CType(Array.CreateInstance(GetType(Byte), 11), Byte())
                            Index(0) = &H81 ' RECORD_VSH / First record of frame
                            Index(1) = 0 ' No start offset information
                            Index(2) = 0 ' No payload size information
                            ' PCR transformation - Starting at 0
                            Dim PCR_Int As Long = CLng(Math.Truncate((Last_PCR - First_PCR)))
                            Index(3) = CByte(PCR_Int And CLng(&HFF))
                            Index(4) = CByte((PCR_Int And CLng(&HFF00)) >> 8)
                            Index(5) = CByte((PCR_Int And CLng(&HFF0000)) >> 16)
                            Dim PCR_Frac As Byte = CByte(((Last_PCR - First_PCR) - CDbl(PCR_Int)) * CDbl(255))
                            Index(6) = PCR_Frac
                            ' TS packet number - Starting at 1
                            Index(7) = CByte(Packets And CLng(&HFF))
                            Index(8) = CByte((Packets And CLng(&HFF00)) >> 8)
                            Index(9) = CByte((Packets And CLng(&HFF0000)) >> 16)
                            Index(10) = CByte((Packets And CLng(&HFF000000)) >> 24)
                            ' Writing the index
                            sw.BaseStream.Write(Index, 0, 11)
                        Else
                            ' We filter this PES (no PCR update)
                            PES_filtered += 1
                        End If
                    End If
                End If

            End If

            If NbRead > 0 And NbRead < 188 Then
                ' Check if we have a uncommon read size
                Console.WriteLine("WARNING: The file doesn't contain an exact number of TS packets!")
            End If
        Loop While NbRead = 188
        sr.Close()
        sw.Close()

        Console.WriteLine(" * " + Packets.ToString + " TS packets were read")
        Console.WriteLine(" * " + PES_Found.ToString + " PES sections were found")
        Console.WriteLine(" * " + PES_filtered.ToString + " PES sections were filtered")
        If (First_PCR > 0 And Last_PCR > 0) Then
            Console.WriteLine(" * Last PCR value is " + Last_PCR.ToString("0.000") + "s")
            Console.WriteLine(" * Movie length estimation is " + (Last_PCR - First_PCR).ToString("0.000") + "s")
        End If
        Console.WriteLine(" * Indexing running time was " + DateTime.Now.Subtract(StartTime).TotalSeconds.ToString("0.000") + "s")
        Console.WriteLine("End of indexing.")

        ' For debugging purpose
        'Console.ReadKey()

    End Sub

End Module
